# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def _install():
    import os

    from tornado import web

    from .._mars.services.web.handlers import handlers, static_handlers
    from .index_handler import XorbitsIndexHandler
    from .version_handler import XorbitsVersionHandler

    handlers["/"] = XorbitsIndexHandler
    static_handlers[r"[^\?\&]*/static/(.*)"] = (
        web.StaticFileHandler,
        {"path": os.path.join(os.path.dirname(__file__), "ui", "static")},
    )
    handlers["/api/xorbits/version"] = XorbitsVersionHandler
